from bs4 import element

from typing import Optional, List


def select_one(tag: element.Tag, selector: str, attribute: str, default: Optional[str] = None) -> Optional[str]:
    el = tag.select_one(selector)
    if el is None:
        return default
    if attribute == 'text':
        return el.text
    return el.get(attribute, default)


def select_text_one(tag: element.Tag, selector: str):
    return select_one(tag, selector, 'text', '')


def select_all(tag: element.Tag, selector: str, attribute: str) -> List[str]:
    result = []
    for el in tag.select(selector):
        if attribute == 'text':
            result.append(el.text)
        else:
            attr_value = el.get(attribute)
            if attr_value is not None:
                result.append(attr_value)
    return result



import asyncio
from abc import ABC, abstractmethod
from typing import Generic, List, Dict, Iterable, Optional, Set

from .client import Client, CrawlerClient
from .typedefs import KT, VT


class Dispatcher(Generic[KT]):
    def __init__(self, state: Iterable[KT]):
        self._done: Set[KT] = set()
        self._new: List[KT] = list(set(state))
        self._all: Set[KT] = set(state)

    def add(self, key: KT):
        if key not in self._all:
            self._all.add(key)
            self._new.append(key)

    def ack(self, key: KT):
        if key in self._all and key not in self._done:
            self._done.add(key)

    def get(self) -> Optional[KT]:
        if len(self._new) != 0:
            return self._new.pop()

    def empty(self) -> bool:
        return len(self._done) == len(self._all)


class Worker(ABC, Generic[KT, VT]):
    @abstractmethod
    async def run(self) -> Dict[KT, VT]:
        """ """


class Master(Generic[KT, VT]):
    def __init__(self, workers: Iterable[Worker[KT, VT]]):
        self._workers = workers

    async def run(self) -> Dict[KT, VT]:
        result = {}
        worker_results = await asyncio.gather(*[
            worker.run() for worker in self._workers
        ])
        for worker_result in worker_results:
            result.update(worker_result)
        return result


class CrawlerWorker(Worker[KT, VT]):
    def __init__(self, dispatcher: Dispatcher[KT], client: CrawlerClient[KT, VT]):
        self._dispatcher = dispatcher
        self._client = client

    async def run(self) -> Dict[KT, VT]:
        results: Dict[KT, VT] = {}
        while not self._dispatcher.empty():
            key = self._dispatcher.get()
            if key is not None:
                result = await self._client.fetch(key)
                if result is not None:
                    new_keys, value = result
                    for new_key in new_keys:
                        self._dispatcher.add(new_key)
                    results[key] = value
                self._dispatcher.ack(key)
            else:
                await asyncio.sleep(1)
        return results


class SimpleWorker(Worker[KT, VT]):
    def __init__(self, dispatcher: Dispatcher[KT], client: Client[KT, VT]):
        self._dispatcher = dispatcher
        self._client = client

    async def run(self) -> Dict[KT, VT]:
        results: Dict[KT, VT] = {}
        while not self._dispatcher.empty():
            key = self._dispatcher.get()
            if key:
                result = await self._client.fetch(key)
                if result is not None:
                    results[key] = result
                self._dispatcher.ack(key)
            else:
                await asyncio.sleep(1)
        return results



import traceback
import aiohttp
from abc import ABC, abstractmethod
from typing import Generic, Optional, Tuple, Iterable

from .cache import Cache
from .typedefs import KT, VT
from .session import SessionPool


class Client(ABC, Generic[KT, VT]):
    @abstractmethod
    async def fetch(self, key: KT) -> Optional[VT]:
        pass


class CrawlerClient(Client, ABC, Generic[KT, VT]):
    @abstractmethod
    async def fetch(self, key: KT) -> Optional[Tuple[Iterable[KT], VT]]:
        pass


class CacheClient(Client[str, VT]):
    def __init__(self, client: Client[str, VT], cache: Cache[str, VT]):
        self._client = client
        self._cache = cache

    async def fetch(self, key: str) -> Optional[VT]:
        cache_value = self._cache.get(key)
        if cache_value is not None:
            return cache_value
        else:
            new_value = await self._client.fetch(key)
            if new_value is not None:
                self._cache.set(key, new_value)
            return new_value


class WebClient(Client[str, bytes]):
    def __init__(self, session_pool: SessionPool):
        self._session_pool = session_pool

    async def fetch(self, url: str) -> Optional[bytes]:
        proxy, session = self._session_pool.rand()
        try:
            response: aiohttp.ClientResponse = await session.get(url, proxy=proxy)
            data = await response.read()
            return data
        except (aiohttp.ClientHttpProxyError, aiohttp.ClientProxyConnectionError):
            if proxy is not None:
                self._session_pool.pop(proxy)
            return None
        except aiohttp.ClientError:
            print(traceback.format_exc())
            return None


class RetryClient(Client[KT, VT]):
    def __init__(self, client: Client, retry_count: int):
        self._retry_count = retry_count
        self._client = client

    async def fetch(self, key: KT) -> Optional[VT]:
        for i in range(self._retry_count):
            result = await self._client.fetch(key)
            if result is not None:
                return result
        return None


class FakeClient(Client[str, str]):
    async def fetch(self, key: str) -> str:
        return key
